{-# LANGUAGE FlexibleContexts #-}
module Ex06 where

import Control.Monad.State
import Control.Monad.Trans
import System.IO
import System.Random
import Test.QuickCheck
import Text.Read (readMaybe)

-- TASK 1 --

-- Implement an IO procedure that reads the given input
-- file, throws away all the lines in the file that
-- appear more than once, and writes the remaining
-- (unique) lines to the given output file.
--
-- You will want to use the `readFile` and `writeFile`
-- procedures.


isUnique :: Eq a => a -> [a] -> Bool
isUnique x ls = length (filter (\y -> y == x) ls) == 1

unique :: Eq a => [a] -> [a]
unique ls = filter (\x -> isUnique x ls) ls

onlyUnique :: FilePath -> FilePath -> IO ()
onlyUnique inputFile outputFile = do
    content <- readFile inputFile
    let ls = lines content
    writeFile outputFile $ unlines $ unique ls

-- TASK 2 --

-- There is a well-known guessing game where the player
-- attempts to guess a secret integer number that has been chosen
-- randomly between 1 and 100.

-- If the player guesses incorrectly, the only information
-- they receive is whether the secret number is lower or
-- higher than the player's guess.

-- The player wins if they successfully guess the secret.
-- If the player fails to guess the secret number in a certain
-- number of turns, they lose.

-- We'll implement this game in Haskell below. Your task
-- will be to develop and test an AI player for this game.


data GameState =
  GameState {
    remainingTurns :: Int,
    secretNumber :: Int
  }
data Outcome = PlayerWon | PlayerLost deriving (Eq, Show)
data Response = Lower | Higher deriving (Eq, Show)

-- Player implementations will be modelled using two
-- "callback" functions: `guess`, which returns the player's
-- next guess, and `signalWrong`, which lets the player know
-- the response (lower or higher) to their last guess.

-- Some players, such as humans, will work in the IO monad,
-- since to get their guess, we need to do input/output.
-- The AI player will work in the State monad: it will make
-- its `guess` using a deterministic knowledge state, and it
-- will update its knowledge state based on responses.
data Player m =
  Player {
    guess :: m Int,
    signalWrong :: Response -> m ()
  }


-- The gameLoop can work in any monad, including m = IO.
gameLoop :: (Monad m) => GameState -> Player m -> m Outcome
gameLoop (GameState 0 _) player = return PlayerLost
gameLoop (GameState r s) player = do
  g <- guess player
  case compare s g of
    LT -> do
      signalWrong player Lower
      gameLoop (GameState (r-1) s) player
    GT -> do
      signalWrong player Higher
      gameLoop (GameState (r-1) s) player
    EQ -> return PlayerWon

-- The Human Player:

human :: Player IO
human = Player { guess = getGuess, signalWrong = printWrong }
  where
    getGuess = do
      putStr "Enter a number (1-100): "
      x <- getLine
      case readMaybe x of
        Nothing -> getGuess
        Just i  -> return i
    printWrong Lower  = putStrLn "Too high!"
    printWrong Higher = putStrLn "Too low!"

play :: IO ()
play = do
  putStrLn "I chose a random number between 1 and 100."
  putStrLn "You have five turns to guess it!"
  s <- randomRIO (1,100)
  outcome <- gameLoop (GameState 5 s) human
  case outcome of
    PlayerWon  -> putStrLn "You got it!"
    PlayerLost -> do
      putStrLn "You ran out of guesses!"
      putStrLn $ "The number was " ++ show s ++ "."


-- The AI player:

data AiState =
  AiState {
    lowerBound :: Int,
    upperBound :: Int
  } deriving (Show, Eq)

midpoint l u = (l+u) `div` 2

ai :: Player (State AiState)
ai = Player { guess = aiGuess, signalWrong = aiWrong } where
  aiGuess :: State AiState Int
  aiGuess = do
    AiState l u <- get
    return $ midpoint l u
  
  aiWrong :: Response -> State AiState ()
  aiWrong Lower = do
    AiState l u <- get
    put $ AiState l (midpoint l u - 1)
  aiWrong Higher = do
    AiState l u <- get
    put $ AiState (midpoint l u + 1) u

{-
As promised, here is an implementation using modify.

I'm paranoid someone is going to have the same problem with
stack and importing modify so I've commented it out, but
this does work.

As you can see, doing this with modify is very nice and clean looking.
This is a really clean way of solving this sort of problem and
indicative of how one often writes stateful code in practice.

I would've liked to have shown this in the lecture, but alas, 
such is stack.

aiWrongModifier :: Response -> AiState -> AiState
aiWrongModifier Lower (AiState l u) = AiState l (midpoint l u - 1)
aiWrongModifier Higher (AiState l u) = AiState (midpoint l u + 1) u

aiWrongModify :: Response -> State AiState ()
aiWrongModify r = modify $ aiWrongModifier r


You can test that it works using the following ai implementation
aiMod :: Player (State AiState)
aiMod = Player { guess = aiGuess, signalWrong = aiWrong } where
  aiGuess :: State AiState Int
  aiGuess = do
    AiState l u <- get
    return $ midpoint l u
  
  aiWrong :: Response -> State AiState ()
  aiWrong = aiWrongModify
-}

-- Testing the AI:

-- A wrapper class so that QuickCheck can
-- generate secret numbers between 1 and 100.
data Secret = Secret Int deriving (Show)
instance Arbitrary Secret where
  arbitrary = Secret <$> choose (1,100)

-- any working AI should be able to win in 100 guesses
prop_winsEventually (Secret s) =
  evalState (gameLoop (GameState 100 s) ai) (AiState 1 100) == PlayerWon

-- a perfect AI can win any game in at most 7 guesses
prop_winsOptimally (Secret s) =
  evalState (gameLoop (GameState 7 s) ai) (AiState 1 100) == PlayerWon

