module Main where

import Data.List
import System.Random
import Control.Monad.State

---------------------------------------------------------
-- PART 1: General Monads -------------------------------
---------------------------------------------------------

-- Recall that previously we used the list monad to
-- enumerate all the possible outcomes of rolling two dice
-- and adding the result.

-- We now want to write a program that can:
-- 1. simulate such dice rolls using random numbers
-- 2. enumerate all possible outcomes

-- We first declare a data type that allows us to specify
-- different dice rolling schemes:
-- [RollD 6, RollD 6, Add 2] would specify 2d6+2, i.e.
-- [RollD 3, RollD 7]
-- roll two six-sided dice, add them up, then add 2 to
-- the result.

data Input
  = RollD Integer
  | Add Integer


-- enumerate gives us a list of all results
enumerate :: [Input] -> [Integer]
enumerate [] = return 0
enumerate (RollD n:xs) = do
    roll <- [1..n]
    rest <- enumerate xs
    return $ roll + rest
enumerate (Add n:xs) = do
    rest <- enumerate xs
    return $ n + rest

-- simulate uses random numbers to give us one random result
simulate :: [Input] -> IO Integer
simulate [] = return 0
simulate (RollD n:xs) = do
    roll <- randomRIO (1, n)
    rest <- simulate xs
    return $ roll + rest
simulate (Add n:xs) = do
    rest <- simulate xs
    return $ n + rest


-- As you can see, the two programs turn out to be nigh-identical.
-- This is a bad kind of repetition: the logic of the algorithm 
-- is, and always will be, the same in both cases. 
-- Monads allow us to resolve this. We will implement a version of
-- the code that works in a generic monad:

calculate :: (Monad m) => ([Integer] -> m Integer) -> [Input] -> m Integer
calculate choose [] = return 0
calculate choose (RollD n:xs) = do
    roll <- choose [1..n]
    rest <- calculate choose xs
    return $ roll + rest
calculate choose (Add n:xs) = do
    rest <- calculate choose xs
    return $ n + rest
 
-- Write a choose function that could be used for enumerate
-- I.e. such that `calculate enumerator` is the same as `enumerate`
enumerator :: [Integer] -> [Integer]
enumerator = id


-- Write a choose function that could be used for simulate
-- I.e. such that `calculate simulator` is the same as `simulate`
simulator :: [Integer] -> IO Integer
simulator xs = do
    i <- randomRIO (0, length xs - 1)
    return (xs !! i)



-- We can even use this to test our function using our own
-- list of random numbers, to check that it's working
-- correctly.

type TestNumbers = [Integer]

{-
We want a choose function where we can enter
the numbers rolled. For instance, we want something
like:
test = calculate tester [RollD 6, RollD 6, Add 3]
runState test [3,4] gives 3 + 4 + 3 == 10
-}
tester :: [Integer] -> State TestNumbers Integer
tester xs = do
     rs <- get
     case rs of
        [] -> error "No more numbers, buddy"
        (t:ts) -> do
            put ts
            case t `elem` xs of
                False -> error "Out of bounds!"
                True -> return t
{-
Alternate implementation which looks a lot nicer but
produces 1 number higher than you gave it,
e.g. runState test [3,4] now gives 4 + 5 + 3 == 12
for test = calculate tester [RollD 6, RollD 6, Add 3]

If you want to use this implementation, make sure to
change TestNumbers to [Int], either in the declaration
of TestNumbers or in the signature below

tester :: [Integer] -> State TestNumbers Integer
tester xs = do
  rs <- get
  put (tail rs)
  return (xs !! head rs)
-}

test = calculate tester [RollD 6, RollD 6, Add 2, RollD 7, RollD 20, Add 3]


---------------------------------------------------------
-- PART 2: Combining Monads -----------------------------
---------------------------------------------------------

{- If you have further time left over after this, you could
   talk about combining monads. A very nice example
   is reimplementing the RPN calculator without zero padding.
   In that case, you have to handle empty stack errors.
   So you need a monad that combines `State Stack` and the
   `Maybe` monads.

   You can always implement such a monad by hand:
-}

newtype Staybe s a = Staybe (s -> Maybe (s, a))
-- Recall that State s a ~= s -> (s, a)

runStaybe :: Staybe s a -> s -> Maybe (s, a)
runStaybe (Staybe f) s = f s 

instance Monad (Staybe s) where
  return x = Staybe (\s -> Just (s, x))
  -- sb :: Staybe s a == s -> Maybe (s, a)
  -- f :: a -> Staybe s b == a -> (s -> Maybe(s, b))
  Staybe sb >>= f = Staybe $ \s ->
    case sb s of
        Nothing -> Nothing
        Just (s', a) -> runStaybe (f a) s'


{- 
These come for free from Monad, so we can
actually use our monad functions in our Applicative
and Functor instances even though you might expect
to implement functor then applicative then monad!
-}

instance Applicative (Staybe s) where
  pure = return
  f <*> x = do
    f' <- f
    x' <- x
    return (f' x')
instance Functor (Staybe s) where
  fmap f xs = pure f <*> xs

{- explain how this is impelmented -}

pop :: Staybe [Double] Double
pop = Staybe $ \s ->
    case s of
        [] -> Nothing
        (x:s') -> Just (s', x)

push :: Double -> Staybe [Double] ()
push x = Staybe $ \s -> Just ((x:s), ())

-- Pops the first two elements from the stack and adds them
add :: Staybe [Double] ()
add = do
    x <- pop
    y <- pop
    push $ x + y

-- return but for Maybe values
liftS :: Maybe a -> Staybe s a
liftS ma = Staybe $ \s ->
    ma >>= \a ->
    return (s, a)


-- newtype StateT s (m :: * -> *) a = StateT {runStateT :: s -> m (a, s)}
-- StateT :: (s -> m (a, s)) -> StateT s m a

{-
lift :: m a -> StateT s m a
lift ma = StateT $ \s ->
    ma >>= a ->
    return (a, s)
-}

push' :: Double -> StateT [Double] Maybe ()
push' x = do
    s <- get
    put (x:s)

pop' :: StateT [Double] Maybe Double
pop' = do
    s <- get
    case s of
        [] -> lift Nothing
        (x:s') -> put s' >> return x

add' :: StateT [Double] Maybe ()
add' = do
    x <- pop'
    y <- pop'
    push' $ x + y

main :: IO ()
main = putStrLn "hello world"
